#!/bin/bash

USING_CUSTOM_CNF_FILE=false
NEW_CNF_FILE=""

SCRIPT_CONFIG_FILE=user-config.cfg
NUMBER_OF_REPLICAS=""
if [ $# == "3" ]; then
	NEW_CNF_FILE=$1
	USING_CUSTOM_CNF_FILE=true
	let NUMBER_OF_REPLICAS=$3-1 #the master counts as a replica. this is used to calc. the number of slaves.
	SCRIPT_CONFIG_FILE=$2 #benchmarking config. file.	
fi

source $SCRIPT_CONFIG_FILE

echo "STATIC_CONFIG_FILE: $STATIC_CONFIG_FILE"
source $STATIC_CONFIG_FILE

echo ">>>> Number of replicas to be created: $NUMBER_OF_REPLICAS"
echo ">>>> Config file being used: $SCRIPT_CONFIG_FILE"

MYSQL="mysql -u$MYSQL_USERNAME -p$MYSQL_PASSWORD -e"

###################################
# LOGGING / ERROR FUNCTIONS

#Lock file and logging details
#NOTE: The script records completed tasks in the following log file; the file has a secondary function as a lock file.
# Upon completion (successful or insuccessful) of the script the log file will be removed. Prior to this its contents will be emailed to the relevant address below.

HOSTNAME=`hostname`
DATE_TIME=`date +%F@%H-%M` # the current date and time

#diagmail: two arguments; the message to be mailed, the subject line of the email.
# Sends an email to the SUCCESS_MAILTARGET address informing the person of the scripts completion, and appending the contents of the logfile.
function diagmail {
	echo "$1" | mail -s "$2" $SUCCESS_MAILTARGET
}  

# diagfail: one argument: error message to be mailed.
# Emails the appropriate alias to inform them of the failure of the backup, and the reason why (which is provided as an argument).
function diagfail {
	
	echo "$1
	
Logs prior to failure:

`cat $LOGFILE`" | mail -s "[Eval] FAILURE: MySQL evalutaion on $HOSTNAME FAILED at $DATE_TIME" $FAILURE_MAILTARGET 

	rm -f $LOGFILE
	trap - INT TERM EXIT #Remove the traps to the following system calls.
	exit 1
}

# toLog: one argument, text to be written to the log file.
# Pre-pend the function argument with the current date and time, then add it to the log file.
function tolog {
	echo "`date +%F@%H-%M-%S`: $1
	" >> $LOGFILE
}
############################################################

#Trap system calls related to the scripts premature exit. Prior to this exit ensure the log file's contents are emailed to the appropriate person.
trap "tolog 'Script unexpectedly called an OS exit command. Performing clean shutdown.';diagfail 'Script unexpectedly quit.'; exit" INT TERM EXIT
#Create a log file
touch $LOGFILE
echo "" > $LOGFILE #ensure it is empty.
tolog "Log file created at $LOGFILE"


################################################################

####################################################################################################
#Start of main script

echo 'Stopping existing MySQL instance.'
/etc/init.d/mysqld stop 1> /dev/null


#Remove any existing log files
echo 'Removing existing MySQL binary log files.'
rm -f /var/lib/mysql/mysql-bin.*

#Install clean MySQL Version:
echo 'Removing existing MySQL installation.'
yum -q remove mysql 1> /dev/null
echo 'Re-installing MySQL.'
yum -q install mysql-server 1> /dev/null
yum -q install mysql-devel 1> /dev/null

echo 'Starting MySQL.'
/etc/init.d/mysqld start 1> /dev/null

if [ $? -ne 0 ]; then
		diagfail "Attempt to start new MySQL instance failed."
		exit 1
	else		
		tolog "Successfully started and installed MySQL instance on $HOSTNAME"
fi

#Set the initial root password:
echo "Setting MySQL $MYSQL_USERNAME password."
/usr/bin/mysqladmin -s -u $MYSQL_USERNAME password "$MYSQL_PASSWORD" #remove -s for debugging.


####################################################################################################
# Update my.cnf config file::

if [ $USING_CUSTOM_CNF_FILE == true ] ; then
	echo "Replacing standard config file for $NEW_CNF_FILE."
	cp -f $MYSQL_CONFIG_FILE $MYSQL_CONFIG_FILE.old
	cp -f $NEW_CNF_FILE $MYSQL_CONFIG_FILE
	tolog "Replaced standard config file for $NEW_CNF_FILE."
fi

echo "Adding master server information to my.cnf file."

#Remove existing replication information from the config file.
for i in "${TO_REMOVE[@]}"
do
    sed -i "s/^$i.*//" "$MYSQL_CONFIG_FILE"
done

#Add new replication information to the config file.
REGEX="compute-0-([0-9][0-9]?).local" #To match the hostname and get a unique number from it.

ID="-1" #It will be changed from this.
#Get the ID for this MySQL instance based on the hostname of the server.
if [[ "$HOSTNAME" =~ "$REGEX" ]]
then
	n=${#BASH_REMATCH[*]}
	ID="${BASH_REMATCH[1]}"
else
	echo "WARNING: didn't find a match on the hostname '$HOSTNAME'"
fi

ID_LINE="server-id=$ID"
LOG_BIN_LINE="log-bin=mysql-bin"
DATABASE_TO_REPLICATE_LINE="binlog-do-db=$DATABASE_TO_REPLICATE"

#write this information to the config file:
sed -i "s/\[mysqld\]/\[mysqld\]\n\r$LOG_BIN_LINE\n\r$DATABASE_TO_REPLICATE_LINE\n\r$ID_LINE\n\r/" "$MYSQL_CONFIG_FILE"

if [ $? -ne 0 ]; then
		diagfail "Attempt to write to config file  on $HOSTNAME failed."
		exit 1
	else		
		tolog "Successfully wrote to config file on $HOSTNAME"
fi





#Restart the MySQL instance, then create the necessary replication users...
/etc/init.d/mysqld restart > /dev/null

if [ $? -ne 0 ]; then
		
	if [ $USE_PKILL == true ] ; then
		pkill -9 -f '.*mysqld.*'
		/etc/init.d/mysqld restart > /dev/null
		
		if [ $? -ne 0 ]; then
			diagfail "Attempt to restart MySQL on $HOSTNAME failed."
			exit 1
		fi
	else 
		diagfail "Attempt to restart MySQL on $HOSTNAME failed."
		exit 1
	fi
else		
		tolog "Successfully restarted MySQL on $HOSTNAME"
fi

####################################################################################################
#Create replication user:

echo "Creating replication user."

`$MYSQL "DROP DATABASE IF EXISTS $DATABASE_TO_REPLICATE;"` #clean any previous install.
`$MYSQL "CREATE DATABASE $DATABASE_TO_REPLICATE;"`

`$MYSQL "DROP USER '$REPLICATION_USER'@'%';"` #clean any previous install.
`$MYSQL "CREATE USER '$REPLICATION_USER'@'%' IDENTIFIED BY '$REPLICATION_PASSWORD';"`

`$MYSQL "GRANT REPLICATION SLAVE ON *.* TO '$REPLICATION_USER'@'%';"`
`$MYSQL "FLUSH TABLES WITH READ LOCK;"`
MASTER_STATUS=`$MYSQL "SHOW MASTER STATUS;"`

#Release locks on tables now that the log file position has been established.
`$MYSQL "UNLOCK TABLES;"`

####################################################################################################
#Parse the master status:

echo "Parsing master status."

MASTER_REGEX=".*(mysql-bin.[0-9]{6})\s*([0-9]+).*"   #extract the bin name and log position from master status.

MYSQL_BIN_NAME=""
MYSQL_LOG_POS=""

#Example master status output: 
#File   Position        Binlog_Do_DB    Binlog_Ignore_DB
#mysql-bin.000004        304     benchmarksql

if [[ "$MASTER_STATUS" =~ "$MASTER_REGEX" ]]
then
	MYSQL_BIN_NAME="${BASH_REMATCH[1]}"
	MYSQL_LOG_POS="${BASH_REMATCH[2]}"
	tolog "Successfully found MySQL log-bin name and log position: $MYSQL_BIN_NAME, $MYSQL_LOG_POS."
else
	diagfail "Couldn't obtain the binary name and log position from the data returned by SHOW MASTER STATUS: $MASTER_STATUS"
	exit 1
fi

####################################################################################################
# Create slave instance(s):
tolog "Adding MySQL slave config data to: ${ALL_REMOTE_INSTANCES[@]:0:$NUMBER_OF_REPLICAS}"

for SLAVE_INSTANCE in "${ALL_REMOTE_INSTANCES[@]:0:$NUMBER_OF_REPLICAS}"
do
	echo "Starting slave instance on $SLAVE_INSTANCE."

	( # Start of subshell
		
		/usr/bin/scp -q $SLAVE_SCRIPT_FILENAME $SLAVE_INSTANCE:$DEST_LOCATION
		
		if [ $? -ne 0 ]; then 
			diagfail "Failed to copy script ($SLAVE_SCRIPT_FILENAME) to $SLAVE_INSTANCE using SCP. Error code: $?"
			exit 1
		fi	
		
		/usr/bin/scp -q $STATIC_CONFIG_FILE $SLAVE_INSTANCE:$DEST_LOCATION
		/usr/bin/scp -q $SCRIPT_CONFIG_FILE $SLAVE_INSTANCE:$DEST_LOCATION

		
		if [ $USING_CUSTOM_CNF_FILE == true ] ; then
			ssh root@$SLAVE_INSTANCE "cp -f $MYSQL_CONFIG_FILE $MYSQL_CONFIG_FILE.old"
			/usr/bin/scp -q $NEW_CNF_FILE $SLAVE_INSTANCE:$MYSQL_CONFIG_FILE
		fi
		
		CONFIG_FILE_NAME=${SCRIPT_CONFIG_FILE##*/}
		
		ssh root@$SLAVE_INSTANCE "/bin/bash $DEST_LOCATION$SLAVE_SCRIPT_FILENAME $HOSTNAME $MYSQL_BIN_NAME $MYSQL_LOG_POS $CONFIG_FILE_NAME"
		
		if [ $? -ne 0 ]; then 
			diagfail "Attempt to create MySQL slave on $SLAVE_INSTANCE failed."
			exit 1
		else		
			tolog "Successfully created MySQL slave on $SLAVE_INSTANCE"
		fi		
	) &
done

####################################################################################################
# Remove config information on instances that are not intended to have slaves.

LENGTH="${#ALL_REMOTE_INSTANCES[@]}"
for TO_REMOVE_INSTANCE in "${ALL_REMOTE_INSTANCES[@]:$NUMBER_OF_REPLICAS:$LENGTH}"
do
	echo "Removing slave config information on $TO_REMOVE_INSTANCE."

	( # Start of subshell
		
		/usr/bin/scp -q $REMOVE_SCRIPT_FILENAME $TO_REMOVE_INSTANCE:$DEST_LOCATION
		
		if [ $? -ne 0 ]; then 
			diagfail "Failed to copy script ($REMOVE_SCRIPT_FILENAME) to $SLAVE_INSTANCE using SCP. Error code: $?"
			exit 1
		fi	
		
		/usr/bin/scp -q $STATIC_CONFIG_FILE $TO_REMOVE_INSTANCE:$DEST_LOCATION

		ssh root@$TO_REMOVE_INSTANCE "/bin/bash $DEST_LOCATION$REMOVE_SCRIPT_FILENAME"
		
		if [ $? -ne 0 ]; then 
			diagfail "Attempt to remove slave information from a remote MySQL config file on $TO_REMOVE_INSTANCE failed."
			exit 1
		else		
			tolog "Successfully removed MySQL slave information on $TO_REMOVE_INSTANCE"
	fi		
	) &
done	
	
	
echo "Waiting for operations on remote instances to complete."

wait #wait until the co-routines finish

echo "Operations on remote instances have completed."

####################################################################################################
#Email success-mailtarget on successful completion of script.

LIST_OF_ALL_REMOTE_INSTANCES=${ALL_REMOTE_INSTANCES[@]:0:$NUMBER_OF_REPLICAS}

if [ -f $LOGFILE ]; then
	diagmail "A replicated MySQL service was successfully installed at `date +%F@%H-%M-%S`.

The master for the service is located on: $HOSTNAME

Slave instances for the service are located at: $LIST_OF_ALL_REMOTE_INSTANCES

Logs of this script's execution:

`cat $LOGFILE`" "[Eval] SUCCESS: MySQL cluster successfully installed."

	trap - INT TERM EXIT #Remove the traps to the following system calls.

	echo "Script has successfully terminated."
fi
