#!/bin/bash

USING_CUSTOM_CNF_FILE=false
NEW_CNF_FILE=""

SCRIPT_CONFIG_FILE=default-config.cfg
SCRIPT_CONFIG_FILE_NAME=$SCRIPT_CONFIG_FILE
NUMBER_OF_REPLICAS=""
if [ $# == "4" ]; then
	NEW_CNF_FILE=$1
	USING_CUSTOM_CNF_FILE=true
  
	SCRIPT_CONFIG_FILE=$2 #benchmarking config. file.	
	SCRIPT_CONFIG_FILE_NAME=${SCRIPT_CONFIG_FILE##*/}
	NUMBER_OF_REPLICAS=$3
	FULL_PATH_TO_DIRECTORY=$4 #this is run inside an su command, so we must change the current directory for the remaining paths in this script to work.
	
	pushd $FULL_PATH_TO_DIRECTORY
fi

STATIC_CONFIG_FILE=static-config.cfg


echo ">>>> Number of replicas to be created: $NUMBER_OF_REPLICAS"
echo ">>>> Config file being used: $SCRIPT_CONFIG_FILE"


source $STATIC_CONFIG_FILE


if [ $? -ne 0 ]; then
	echo "> Failed to read config file '$STATIC_CONFIG_FILE'. Do you have permission to access this file?"
	exit 1
else		
	echo "> Successfully read config parameters."
fi
	
source $SCRIPT_CONFIG_FILE	

LOAD_BALANCER_NODE=${ALL_REMOTE_INSTANCES[@]:0:1}
REPLICATION_SERVER_NODE=${ALL_REMOTE_INSTANCES[@]:1:1}

FIRST_DATA_NODE=$HOSTNAME

let MINUS_ONE=$NUMBER_OF_REPLICAS-1
DATA_NODES=("${ALL_REMOTE_INSTANCES[@]:2:$MINUS_ONE}")

echo "Load balancer node: $LOAD_BALANCER_NODE"

echo "Replication server node: $REPLICATION_SERVER_NODE"

echo "Data nodes: ${DATA_NODES[@]} $FIRST_DATA_NODE"

touch $LOGFILE	

if [ $? -ne 0 ]; then
	echo "> Failed to create logfile at $LOGFILE. Do you have permission to create/access this file?"
	exit 1
else		
	echo "> Successfully created log file."
fi


#This script is divided as follows:
#0. Utility and logging functions used through the rest of the script.
#... start of script execution:
#1. Kill existing instances on each machine >> this operation uses pkill to ensure that all instances are killed, as extant instances can affect the rest of the install process.
#2. Install pgcluster on all instances >> connect to each instance and, if pgcluster is not already installed, install it (this takes 20 minutes). If it is installed, delete the existing database, and recreate it.
#3. PGCluster should now be installed >> create the config files needed on each machine to set up the cluster with the correct configuration of data nodes, load balancers, and replication nodes.
#4. Start each instance in sequence >> first start data nodes at once, then wait 7 seconds (enough time for the data nodes to start) and start the replication server. Wait again and start the load balancer.
#5. Create a database in the new system >> this is used partly to test that the database system was successfully created and started. 
#6. Email the 'success-mailtarget' on successful completion of script.

###################################
#0. Utility and logging functions used through the rest of the script.

HOSTNAME=`hostname`
DATE_TIME=`date +%F@%H-%M` # the current date and time



#diagmail: two arguments; the message to be mailed, the subject line of the email.
# Sends an email to the SUCCESS_MAILTARGET address informing the person of the scripts completion, and appending the contents of the logfile.
function diagmail {
	echo "> $1" | mail -s "$2" $SUCCESS_MAILTARGET
}  

# diagfail: one argument: error message to be mailed.
# Emails the appropriate alias to inform them of the failure of the backup, and the reason why (which is provided as an argument).
function diagfail {
	
	echo "> $1
	
Logs prior to failure:

`cat $LOGFILE`" | mail -s "[Eval] FAILURE: PGCluster Installation on $HOSTNAME failed at $DATE_TIME" $FAILURE_MAILTARGET 

	rm -f $LOGFILE
	trap - INT TERM EXIT #Remove the traps to the following system calls.
	exit 1
}

# toLog: one argument, text to be written to the log file.
# Pre-pend the function argument with the current date and time, then add it to the log file.
function tolog {
	echo "> `date +%F@%H-%M-%S`: $1
	" >> $LOGFILE
}
############################################################

#Trap system calls related to the scripts premature exit. Prior to this exit ensure the log file's contents are emailed to the appropriate person.
trap "tolog 'Script unexpectedly called an OS exit command. Performing clean shutdown.';diagfail 'Script unexpectedly quit.'; exit" INT TERM EXIT
#Create a log file
touch $LOGFILE
echo "> " > $LOGFILE #ensure it is empty.
tolog "Log file created at $LOGFILE"


################################################################

DEST_LOCATION="/tmp"

function runScriptRemoteAsync { #$1 - script to run $2 - node to run it on, $3 - the hostname of the load balancer node.
	( # Start of subshell
			echo "> Copying then running $1 on $2: /usr/bin/scp -q $1 $NON_ROOT_USER@$2:$DEST_LOCATION"
			
			/usr/bin/scp -q $1 $NON_ROOT_USER@$2:$DEST_LOCATION
			
			if [ $? -ne 0 ]; then 
				diagfail "Failed to copy script ($1) to $2 using SCP: /usr/bin/scp -q $1 $NON_ROOT_USER@$2:$DEST_LOCATION"
				exit 1
			fi	
			
			
			
			/usr/bin/scp  $STATIC_CONFIG_FILE $NON_ROOT_USER@$2:$DEST_LOCATION	#copy the config files accross as well.
			echo "/usr/bin/scp  $STATIC_CONFIG_FILE $NON_ROOT_USER@$2:$DEST_LOCATION"
			/usr/bin/scp  $SCRIPT_CONFIG_FILE $NON_ROOT_USER@$2:$DEST_LOCATION
			
			if [ $? -ne 0 ]; then 
				diagfail "Failed to copy config file to $2 using SCP. Error code: $?"
				exit 1
			fi	
			
			ssh $NON_ROOT_USER@$2 "/bin/bash $DEST_LOCATION/$1 $DEST_LOCATION/$SCRIPT_CONFIG_FILE_NAME $3 $NUMBER_OF_REPLICAS $4"
		echo "ssh $NON_ROOT_USER@$2 /bin/bash $DEST_LOCATION/$1 $DEST_LOCATION/$SCRIPT_CONFIG_FILE_NAME $3 $NUMBER_OF_REPLICAS $4"	
			if [ $? -ne 0 ]; then 
				diagfail "Attempt to run $1 on $2 failed."
				exit 1
			else		
				tolog "Successfully run $1 on $2"
			fi		
		) &
}

function suExecuteCommand { #$1 - command to execute, $2 location to execute, $3 - user to execute as. $4=false if no error checking (i.e. it is acceptable if the command fails).
	ssh $NON_ROOT_USER@$2 "$1"
	
	if [ $? -ne 0 ]; then
		
		if [ "$4" = "false" ]; then
			echo "> Failed to excute command '$1' on '$2' as '$3' user (though this is expected)."
		else
			diagfail "Failed to excute command '$1' on '$2' as '$3' user"
			exit 1
		fi
	else		
		tolog "Successfully executed '$1' on '$2' as '$3'"
	fi
}

#SCP the psqlrc config file and the postgres.conf file to each machine.
function transferConfigFiles { #$1 - host to transfer file to...
	ssh $NON_ROOT_USER@$1 "cp -f $POSTGRES_CONFIG_FILE $POSTGRES_CONFIG_FILE.old"
	/usr/bin/scp -q $NEW_CNF_FILE $NON_ROOT_USER@$1:$POSTGRES_CONFIG_FILE
	
	if [ $? -ne 0 ]; then
		
		echo "Failed to excute command: /usr/bin/scp -q $NEW_CNF_FILE $NON_ROOT_USER@$1:$POSTGRES_CONFIG_FILE"
		diagfail "Failed to excute command: /usr/bin/scp -q $NEW_CNF_FILE $NON_ROOT_USER@$1:$POSTGRES_CONFIG_FILE"
		exit 1

	fi
	
	ssh $NON_ROOT_USER@$1 "mkdir -p $PG_ROOT/etc; cp -f $PSQLRC_FILE $PSQLRC_FILE.old"
	
	if [ $AUTO_COMMIT_ON == true ] ; then
		/usr/bin/scp -q $NEW_PSQLRC_FILE_ON $NON_ROOT_USER@$1:$PSQLRC_FILE
	else
		/usr/bin/scp -q $NEW_PSQLRC_FILE_OFF $NON_ROOT_USER@$1:$PSQLRC_FILE
	fi	
}

####################################################################################################
######### Start of script execution.....
####################################################################################################

UNZIP_PATH=/tmp
PATH_TO_PGCLUSTER=$UNZIP_PATH/pgcluster-1.9.0rc7

##############################################################################
#1. Kill existing instances on each machine >> this operation uses pkill to ensure that all instances are killed, as extant instances can affect the rest of the install process.

echo "> Terminating any extant pgcluster processes. Error messages may be shown if none are running."
echo "> Non-root user: $NON_ROOT_USER"
for DATA_NODE in "${DATA_NODES[@]}"
do
echo "> About to execute..."
	
	ssh $NON_ROOT_USER@$DATA_NODE "/usr/local/pgsql/bin/pg_ctl -D $PG_DATA_ROOT -w stop > /dev/null" > /dev/null
	echo "$NON_ROOT_USER@$DATA_NODE \"/usr/local/pgsql/bin/pg_ctl -D $PG_DATA_ROOT -w stop > /dev/null\" > /dev/null"
done

ssh $NON_ROOT_USER@$FIRST_DATA_NODE "/usr/local/pgsql/bin/pg_ctl -D $PG_DATA_ROOT -w stop > /dev/null" > /dev/null
	
echo "ssh $NON_ROOT_USER@$REPLICATION_SERVER_NODE /usr/local/pgsql/bin/pgreplicate -D /usr/local/pgsql/data stop > /dev/null"
ssh $NON_ROOT_USER@$REPLICATION_SERVER_NODE "/usr/local/pgsql/bin/pgreplicate -D /usr/local/pgsql/data stop > /dev/null"
ssh $NON_ROOT_USER@$LOAD_BALANCER_NODE "/usr/local/pgsql/bin/pglb -D /usr/local/pgsql/data stop" > /dev/null

#pkill everything.
pkill -9 -f '.*$PG_ROOT/.*'

for DATA_NODE in "${ALL_REMOTE_INSTANCES[@]}"
do
echo "ssh $NON_ROOT_USER@$DATA_NODE 'pkill -9 -f '.*$PG_ROOT/.*' > /dev/null'"
ssh $NON_ROOT_USER@$DATA_NODE "pkill -9 -f '.*$PG_ROOT/.*' > /dev/null"
	
	
done

##############################################################################
#2. Install pgcluster on all instances >> connect to each instance and, if pgcluster is not already installed, install it (this takes 20 minutes). If it is installed, delete the existing database, and recreate it.

ALL_INSTANCES_IN_USE=("${DATA_NODES[@]}" $REPLICATION_SERVER_NODE $LOAD_BALANCER_NODE $FIRST_DATA_NODE)

tolog "About to install PGCluster on machines: ${ALL_INSTANCES_IN_USE[@]}"


for PG_INSTANCE in "${ALL_INSTANCES_IN_USE[@]}"
do
	( # Start of subshell
		echo "> Executing: ssh $NON_ROOT_USER@$PG_INSTANCE \"test -d $PG_ROOT && echo true\""
		if [ `ssh $NON_ROOT_USER@$PG_INSTANCE "test -d $PG_ROOT && echo true"` == "true" ]; then 
			#directory exists
			echo "> PGCluster already installed on $PG_INSTANCE. Creating database at $PG_DATA_ROOT."
			ssh $NON_ROOT_USER@$PG_INSTANCE "rm -rf  $PG_DATA_ROOT" 1> /dev/null
			ssh $NON_ROOT_USER@$PG_INSTANCE "mkdir $PG_DATA_ROOT" 1> /dev/null
			 echo "ssh $NON_ROOT_USER@$PG_INSTANCE mkdir $PG_DATA_ROOT 1> /dev/null"
			ssh $NON_ROOT_USER@$PG_INSTANCE "chown -R $NON_ROOT_USER $PG_ROOT;" > /dev/null
			
			ssh $NON_ROOT_USER@$PG_INSTANCE "$PG_ROOT/bin/initdb -D $PG_DATA_ROOT > /dev/null" > /dev/null

			if [ $? -ne 0 ]; then
				echo "> Failed to initialize database on $PG_INSTANCE."
				exit 1
			else		
				echo "> Finished initializing database on $PG_INSTANCE."
			fi			
		else 
			#if the pgsql directory doesn't already exist, install								
			ZIP_FILE_NAME=$UNZIP_PATH/pgcluster.tar.gz
			
			echo "> Installing PGCluster on $PG_INSTANCE. This normally takes 20 minutes (to build and install)."
			ssh $NON_ROOT_USER@$PG_INSTANCE "wget -q -O $ZIP_FILE_NAME $PG_CLUSTER_URL"
			 
			ssh $NON_ROOT_USER@$PG_INSTANCE "tar -zxf $ZIP_FILE_NAME -C $UNZIP_PATH; rm -f $ZIP_FILE_NAME 1> /dev/null"

			ssh $NON_ROOT_USER@$PG_INSTANCE "pushd $PATH_TO_PGCLUSTER; bash $PATH_TO_PGCLUSTER/configure 1> /dev/null; popd;"
			
			if [ $? -ne 0 ]; then
				echo "> Failed to configure PGCluster install on $PG_INSTANCE."
				exit 1
			else		
				echo "> Finished configuring build on $PG_INSTANCE."
			fi
			
			#It's possible that the the following gmake operations fail (i.e. return an error code), but the install was successful, so the script doesn't use the error code to check for failure.
			echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>"
			echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>"
			echo "> IMPORTANT::"
			echo "> Started 'gmake' on $PG_INSTANCE at `date +%F@%H-%M`. This will take around 20 minutes on most machines."
			echo "> In most cases several warnings will appear, though installation can be successful regardless. The only way of testing this is to create a database, which this script does later on."
			echo "> You may be required to enter your root password at this point."
			echo "<<<<<<<<<<<<<<<<<<<<<<<<<<"
		echo "pushd $PATH_TO_PGCLUSTER; gmake > /dev/null; gmake install > /dev/null; popd; mkdir $PG_DATA_ROOT; adduser $NON_ROOT_USER > /dev/null; chown -R $NON_ROOT_USER $PG_ROOT > /dev/null;"
	
			ssh $ROOT_USER@$PG_INSTANCE "pushd $PATH_TO_PGCLUSTER; gmake > /dev/null; gmake install > /dev/null; popd; mkdir $PG_DATA_ROOT; adduser $NON_ROOT_USER > /dev/null; chown -R $NON_ROOT_USER $PG_ROOT > /dev/null;" 
			echo "> Finished 'gmake' on $PG_INSTANCE."
			
	
			 
			ssh $NON_ROOT_USER@$PG_INSTANCE "$PG_ROOT/bin/initdb -D $PG_DATA_ROOT" 1> /dev/null
			
			if [ $? -ne 0 ]; then
				echo "> Failed to initiate PGCluster instance on $PG_INSTANCE."
				exit 1
			else		
				echo "> Successfully initiated PGCluster instance on $PG_INSTANCE"
			fi
			 
		fi
	) &
	
done

wait #wait until installed on every instance.

####################################################################################################
#3. PGCluster should now be installed >> create the config files needed on each machine to set up the cluster with the correct configuration of data nodes, load balancers, and replication nodes.
tolog "Creating config files at each instance."



for DATA_NODE in "${DATA_NODES[@]}"
do
	transferConfigFiles $DATA_NODE
	runScriptRemoteAsync configure-data-instance.sh $DATA_NODE	
done

transferConfigFiles $FIRST_DATA_NODE
runScriptRemoteAsync configure-data-instance.sh $FIRST_DATA_NODE

transferConfigFiles $LOAD_BALANCER_NODE
runScriptRemoteAsync configure-load-balancer.sh $LOAD_BALANCER_NODE $FIRST_DATA_NODE	

transferConfigFiles $REPLICATION_SERVER_NODE
runScriptRemoteAsync configure-replication-instance.sh $REPLICATION_SERVER_NODE	$FIRST_DATA_NODE

echo "Waiting for all scripts to be written."
wait #wait until config files have been written at every location.
echo "All scripts have been written and run."
####################################################################################################
#4. Start each instance in sequence >> first start data nodes at once, then wait 7 seconds (enough time for the data nodes to start) and start the replication server. Wait again and start the load balancer.
#Start instances sequentially...

echo "> Starting data nodes: ${DATA_NODES[@]}."

( suExecuteCommand "$PG_ROOT/bin/pg_ctl -w -D $PG_DATA_ROOT start" $FIRST_DATA_NODE $NON_ROOT_USER ) &

for DATA_NODE in "${DATA_NODES[@]}"
do
	echo "> Starting data node on $DATA_NODE."
	#The following command is run in a separate sub-shell because it doesn't successfully return control to the script (as it should according to the use of the -W parameter). Consequently the script doesn't wait for this line to return with the 'wait' command, but sleeps for as long as it should take to execute.
	( suExecuteCommand "$PG_ROOT/bin/pg_ctl -w -D $PG_DATA_ROOT start" $DATA_NODE $NON_ROOT_USER ) &
	echo "> Started data node on $DATA_NODE."
done

sleep 7

echo "> Starting the replication server."
suExecuteCommand "$PG_ROOT/bin/pgreplicate -D $PG_DATA_ROOT start" $REPLICATION_SERVER_NODE $NON_ROOT_USER

echo "> Starting the load balancer."
( suExecuteCommand "$PG_ROOT/bin/pglb -D $PG_DATA_ROOT start" $LOAD_BALANCER_NODE $NON_ROOT_USER ) &

sleep 7



#5. Create a database in the new system >> this is used partly to test that the database system was successfully created and started. 
echo "> Creating database."

suExecuteCommand "$PG_ROOT/bin/createdb $BENCHMARK_DATABASE" $FIRST_DATA_NODE $NON_ROOT_USER

####################################################################################################
#6. Email the 'success-mailtarget' on successful completion of script.

LIST_OF_DATA_INSTANCES=${DATA_NODES[@]}

if [ -f $LOGFILE ]; then
	diagmail "A replicated PGCluster service was successfully installed at `date +%F@%H-%M-%S`.

The load balancer for the service is located on: $LOAD_BALANCER_NODE

The replication server for the service is located on: $REPLICATION_SERVER_NODE

Data instances for the service are located at: $LIST_OF_DATA_INSTANCES

Logs of this script's execution:

`cat $LOGFILE`" "[Eval] SUCCESS: PGCluster Installed."

	trap - INT TERM EXIT #Remove the traps to the following system calls.

	popd
	
	echo "> Script has successfully terminated."
fi
