#!/bin/bash

if [ $# != "1" ]; then
	echo "You must provide the name of the .benchmark file you wish to run."
	exit 1
fi

echo "Running benchmark configuration: $1"


CONFIG_FILE=$1
if [[ $1 != /* ]]; then #get the full path if not already.
CONFIG_FILE=`pwd`/$1
fi


source $CONFIG_FILE

echo "Config file: $CONFIG_FILE"
MYSQL_DATABASE_NAME="${MYSQL[@]:0:1}"
MYSQL_ARRAY_SIZE=${#MYSQL[@]}
MYSQL_CONFIG_FILES="${MYSQL[@]:1:$MYSQL_ARRAY_SIZE}"

POSTGRES_DATABASE_NAME="${POSTGRES[@]:0:1}"
POSTGRES_ARRAY_SIZE=${#POSTGRES[@]}
POSTGRES_CONFIG_FILES="${POSTGRES[@]:1:$POSTGRES_ARRAY_SIZE}"

H2O_DATABASE_NAME="${H2O[@]:0:1}"

echo "Running benchmarks for:"
echo "$MYSQL_DATABASE_NAME, with configs: $MYSQL_CONFIG_FILES"
echo "$POSTGRES_DATABASE_NAME, with configs: $POSTGRES_CONFIG_FILES"
echo "$H2O_DATABASE_NAME"

echo "Other running Java processes:"
ps aux | grep java

#Execute the given benchmark against the specified ($1) database.
#$1 - name of the folder which the database script files are located; also the name of the database properties file (ie. $1/cnf/$1.properties).
#$2 - name of the configuration to run
function executeAgainstDatabase {
	
	
	#Hack the configuration files into this function because bash can't pass arrays as parameters.
	CONFIGURATIONS=""
	DATABASE_NAME=$1
	
	if [ $DATABASE_NAME == "mysql" ] ; then
		CONFIGURATIONS=$MYSQL_CONFIG_FILES
	elif [ $DATABASE_NAME == "postgresql" ] ; then
		CONFIGURATIONS=$POSTGRES_CONFIG_FILES
	else
		CONFIGURATIONS=("N/A") #it's a database that doesn't make use of different config files.
	fi
	
	CONNECTION_INFO="cnf/$DATABASE_NAME.properties"
	
	
	for NUMBER_OF_REPLICAS in "${NUMBER_OF_REPLICAS_ARRAY[@]}"
	do
		for CONFIG in "${CONFIGURATIONS[@]}"
		do
			
			#Perform actual test runs:	
			for (( c=1; c<=$NUM_OF_RUNS; c++ ))
			do

#				rm -rf /tmp/RESULTS/benchmark-results/2011*	#disk fills up too quickly with full diagnostic.
				#The H2O start script occasionally fails, so the while loop below attempts to create the database and run the benchmark multiple times
				#It stops either at 10 attempts, or when the database is successfully created.
				SUCCESSFULLY_STARTED=false
				COUNTER=10
				while [ $SUCCESSFULLY_STARTED == false ] && [ $COUNTER -gt 0 ]
				do
					COUNTER=$(( $COUNTER - 1 ))
					echo "Running benchmark against '$DATABASE_NAME' with configuration file '$CONFIG'."
				
					FULL_PATH_TO_DIRECTORY=`pwd`/$DATABASE_NAME
				
					pushd $DATABASE_NAME

					if [ $DATABASE_NAME == "postgresql" ] ; then
						#postgresql must be run as non-root user.
						su -c "$FULL_PATH_TO_DIRECTORY/start-cluster.sh $CONFIG $CONFIG_FILE $NUMBER_OF_REPLICAS $FULL_PATH_TO_DIRECTORY" - $NON_ROOT_USER
					else
						./start-cluster.sh $CONFIG $CONFIG_FILE $NUMBER_OF_REPLICAS
					fi
					
					if [ $? -ne 0 ]; then
							echo "ERROR: Failed to execute script to create $DATABASE_NAME system. Does this user have permission to execute this script?"
						popd
						sleep 2
						continue
					fi
					
					sed -i "s/<connectionline>/conn=jdbc:$DATABASE_NAME:\/\/$HOSTNAME:5432\/$BENCHMARK_DATABASE/" "$CONNECTION_INFO" #pgcluster-specific command, but it will not break mysql configs.

					../individual-benchmark-run.sh $CONFIG_FILE $CONFIG $NUMBER_OF_REPLICAS $CONNECTION_INFO $DATABASE_NAME $MINUTES_TO_EXECUTE
					
					if [ $? -ne 0 ]; then
						echo "ERROR: Failed to execute benchmark script: individual-benchmark-run.sh $CONFIG_FILE $CONFIG $NUMBER_OF_REPLICAS $CONNECTION_INFO $DATABASE_NAME"
						sleep 2
						popd
						continue
					else
						SUCCESSFULLY_STARTED=true
					fi
					
					popd
					
				done
								
				#Increment the number of minutes each benchmark takes every $INCREMENT_FREQUENCY times.
				mod=$(($c % $INCREMENT_FREQUENCY))   
				if [ "$mod" -eq 0 ]; then
					MINUTES_TO_EXECUTE=$(($MINUTES_TO_EXECUTE+$INCREMENT_NUMBER))
				fi
			
			done
		done
	done
}

executeAgainstDatabase $H2O_DATABASE_NAME
#executeAgainstDatabase $MYSQL_DATABASE_NAME
#executeAgainstDatabase $POSTGRES_DATABASE_NAME
